import re

import matplotlib.pyplot as plt

import Bio
import Bio.Restriction
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation

import dnacauldron as dc

try:
    from dna_features_viewer import BiopythonTranslator
except ImportError:

    class AssessmentTranslator:
        """Please install dna_features_viewer to use this class."""

        def __init__(self):
            raise Exception("Please install dna_features_viewer to use this class.")

else:

    class AssessmentTranslator(BiopythonTranslator):
        """Custom translator for highlighting key features."""

        def compute_feature_color(self, feature):
            assessment_ref = "seq_assessment"
            if assessment_ref in feature.qualifiers:
                if feature.qualifiers[assessment_ref] == "enzyme":
                    return "red"
                elif feature.qualifiers[assessment_ref] == "excised":
                    return "yellow"
                elif feature.qualifiers[assessment_ref] == "backbone":
                    return "tab:cyan"
                else:
                    return "tab:blue"  # default dna_features_viewer color
            else:
                return "tab:blue"


topology_options = ["linear", "circular"]  # A DNA molecule can be one of these


class Assessment:
    """The sequence assessment class.

    Parameters
    ----------
    record : SeqRecord
        A Biopython `SeqRecord`.
    enzyme : str
        A restriction enzyme. A Biopython `RestrictionType` will be looked up using the string.
    """

    UNKNOWN_IDS = [
        "None",
        "",
        "<unknown id>",
        ".",
        "EXPORTED",
        "<unknown name>",
        "Exported",
    ]

    def __init__(self, record, enzyme):
        self.record = record
        self.record_len = len(record)
        if "topology" not in self.record.annotations:
            raise Exception("Sequence topology not set (use `linear` or `circular`)!")
        if self.record.annotations["topology"] not in topology_options:
            raise Exception("Sequence topology must be `linear` or `circular`!")
        self.topology = self.record.annotations["topology"]
        if self.topology == "linear":
            self.is_linear = True
        else:
            self.is_linear = False
        self.enzyme = Bio.Restriction.__dict__[enzyme]
        self.enzyme_name = str(self.enzyme)
        self.results = {}

    def assess_seq(self, other_enzymes=None):
        """Evaluate sequence for Golden Gate.

        Parameters
        ----------
        other_enzymes : list
            List of enzymes used in higher level assemblies.
        """
        if other_enzymes:
            self.other_enzymes = ", ".join([str(enz) for enz in other_enzymes])
        self.add_name()
        self.get_number_of_sites()
        self.evaluate_orientation()
        self.digest_sequence()
        self.count_other_sites(other_enzymes)
        self.check_enzyme_site_locations()
        self.sum_results()
        self.plot_sequence()

    def add_name(self):
        """Set a name for the assessment."""
        # To display in the report:
        if str(self.record.id).strip() in self.UNKNOWN_IDS:
            self.name = "Unnamed"
        else:
            if len(self.record.id) > 20:  # to fit page width
                self.name = self.record.id[:17] + "..."
            else:
                self.name = self.record.id

    def get_number_of_sites(self):
        restriction_batch = Bio.Restriction.RestrictionBatch([self.enzyme])
        analysis = Bio.Restriction.Analysis(
            restriction_batch, sequence=self.record.seq, linear=self.is_linear
        )
        self.analysis_results = analysis.full(linear=self.is_linear)

        self.results["number_of_sites"] = len(self.analysis_results[self.enzyme])

        # Add as features for plot in report:
        for enzyme, sites in self.analysis_results.items():
            for site in sites:
                self.record.features.append(
                    SeqFeature(
                        FeatureLocation(site, site + 1),
                        id=str(enzyme),
                        type="misc_feature",
                        qualifiers={
                            "label": str(enzyme),
                            "seq_assessment": "enzyme",
                        },
                    )
                )

    def evaluate_orientation(self):
        self.results["is_site_orientation_correct"] = False  # default
        # Forward strand:
        self.iter_forward = [
            match.end() for match in re.finditer(self.enzyme.site, str(self.record.seq))
        ]
        if sum(1 for _ in self.iter_forward) == 1:
            self.forward_enzyme = self.iter_forward[0]
            # rev_complement_site = str(self.record.seq.reverse_complement())
            rev_complement_site = str(Seq(self.enzyme.site).reverse_complement())
            self.iter_reverse = [
                m.start()
                for m in re.finditer(rev_complement_site, str(self.record.seq))
            ]
            if sum(1 for _ in self.iter_reverse) == 1:  # 1 site in both strands:
                self.results["is_site_orientation_correct"] = True
                self.reverse_enzyme = self.iter_reverse[0]

        if self.results["is_site_orientation_correct"]:
            if self.reverse_enzyme < self.forward_enzyme:
                self.record.features.append(
                    SeqFeature(
                        FeatureLocation(
                            self.reverse_enzyme - 1, self.forward_enzyme + 1
                        ),
                        id=str(self.enzyme),
                        type="misc_feature",
                        qualifiers={
                            "label": "Excised",
                            "seq_assessment": "excised",
                        },
                    )
                )
            else:  # put annotation together from two pieces:
                self.record.features.append(
                    SeqFeature(
                        FeatureLocation(0, self.forward_enzyme + 1),
                        id=str(self.enzyme),
                        type="misc_feature",
                        qualifiers={
                            "label": "Excised",
                            "seq_assessment": "excised",
                        },
                    )
                )
                self.record.features.append(
                    SeqFeature(
                        FeatureLocation(self.reverse_enzyme - 1, len(self.record)),
                        id=str(self.enzyme),
                        type="misc_feature",
                        qualifiers={
                            "label": "Excised",
                            "seq_assessment": "excised",
                        },
                    )
                )

    def digest_sequence(self):
        # Obtain fragments and get the backbone's overhangs.
        self.results["digest"] = {}
        if not self.results["is_site_orientation_correct"]:
            return
        self.results["digest"]["excised_seq"] = []
        self.results["digest"]["backbone_seq"] = []
        if self.is_linear:
            # The sequence has two, correctly oriented enzyme sites.
            # Therefore there are three fragments, and one should contain both sites.
            record_fragments = dc.StickyEndFragment.list_from_record_digestion(
                record=self.record, enzyme=self.enzyme, linear=self.is_linear
            )
            for i in [0, 1, 2]:
                if self.enzyme.site in record_fragments[i].to_standard_string():
                    self.results["digest"]["excised_seq"] += [record_fragments[i]]
                else:
                    self.results["digest"]["backbone_seq"] += [record_fragments[i]]
            self.results["digest"]["left_overhang"] = str(
                self.results["digest"]["backbone_seq"][0].seq.left_end
            )
            self.results["digest"]["right_overhang"] = str(
                self.results["digest"]["backbone_seq"][0].seq.right_end
            )  # there should be only one
        else:  # circular, no other case possible
            # The sequence has two, correctly oriented enzyme sites.
            # Therefore there are exactly two fragments, with one containing both sites.
            record_fragments = dc.StickyEndFragment.list_from_record_digestion(
                record=self.record, enzyme=self.enzyme, linear=self.is_linear
            )
            if self.enzyme.site in record_fragments[0].to_standard_string():
                retain_index = 1  # there are only two fragments
                excise_index = 0
            else:
                retain_index = 0
                excise_index = 1  # reversed
            self.results["digest"]["backbone_seq"] += [record_fragments[retain_index]]
            self.results["digest"]["excised_seq"] += [record_fragments[excise_index]]
            self.results["digest"]["left_overhang"] = str(
                record_fragments[retain_index].seq.left_end
            )
            self.results["digest"]["right_overhang"] = str(
                record_fragments[retain_index].seq.right_end
            )

    def count_other_sites(self, other_enzymes):
        self.results["other_sites"] = {}
        self.results["other_sites"]["has_any_other_sites"] = False

        if other_enzymes is None:
            return
        bio_enzymes = [Bio.Restriction.__dict__[enzyme] for enzyme in other_enzymes]

        restriction_batch = Bio.Restriction.RestrictionBatch(bio_enzymes)
        # Work with the assumption that the sequence is circular:
        analysis = Bio.Restriction.Analysis(
            restriction_batch, sequence=self.record.seq, linear=self.is_linear
        )
        self.results["other_sites"]["enzyme"] = analysis.full(linear=self.is_linear)

        for enzyme, matches in self.results["other_sites"]["enzyme"].items():
            if len(matches) != 0:
                self.results["other_sites"]["has_any_other_sites"] = True
                # Also add as features for plot in report:
                for site in matches:
                    self.record.features.append(
                        SeqFeature(
                            FeatureLocation(site, site + 1),
                            id=str(enzyme),
                            type="misc_feature",
                            qualifiers={
                                "label": str(enzyme),
                                "seq_assessment": "enzyme",
                            },
                        )
                    )

    def check_enzyme_site_locations(self):
        """Flag enzyme sites that are within the retained section."""
        try:
            self.results["other_sites"]["has_any_other_sites"]
            self.results["is_site_orientation_correct"]
        except KeyError:
            print("Run assessment methods first!")
        else:
            self.sites_outside_excised_region = {}
            if (
                self.results["other_sites"]["has_any_other_sites"]
                and self.results["is_site_orientation_correct"]
            ):
                # if there are no other sites, no need to run:
                if self.reverse_enzyme < self.forward_enzyme:
                    # orientation = reverse -> forward
                    for enzyme, sites in self.results["other_sites"]["enzyme"].items():
                        problem_sites = []
                        for site in sites:
                            if self.reverse_enzyme < site < self.forward_enzyme:
                                pass
                            else:
                                problem_sites += [str(site)]
                        if problem_sites != []:
                            self.sites_outside_excised_region[str(enzyme)] = (
                                problem_sites
                            )
                    txt = ""  # for the pdf report
                    for (
                        enzyme,
                        problem_sites,
                    ) in self.sites_outside_excised_region.items():
                        txt += enzyme + ": " + " ".join(problem_sites) + ";"
                    self.sites_outside_excised_region_txt = txt
                else:
                    # orientation = forward -> reverse
                    for enzyme, sites in self.results["other_sites"]["enzyme"].items():
                        problem_sites = []
                        for site in sites:
                            if self.forward_enzyme < site < self.reverse_enzyme:
                                # in this case the site is within the retained backbone
                                problem_sites += [str(site)]
                        if problem_sites != []:
                            self.sites_outside_excised_region[str(enzyme)] = (
                                problem_sites
                            )
                    txt = ""  # for the pdf report
                    for (
                        enzyme,
                        problem_sites,
                    ) in self.sites_outside_excised_region.items():
                        txt += enzyme + ": " + " ".join(problem_sites) + ";"
                    self.sites_outside_excised_region_txt = txt

            else:  # no other sites or orientation not correct
                self.sites_outside_excised_region_txt = ""

    def sum_results(self):
        self.results["pass"] = True
        if self.results["is_site_orientation_correct"] is False:
            # implicitly checks number of sites too
            self.results["pass"] = False
            return
        if self.sites_outside_excised_region_txt:
            self.results["pass"] = False
            return
        # if self.results["other_sites"]["has_any_other_sites"]:
        #     self.results["pass"] = False
        #     return

    def plot_sequence(self):
        """Plot an outline of the sequence."""

        fig, ax = plt.subplots(figsize=(7, 4))
        graphic_record = AssessmentTranslator().translate_record(self.record)
        graphic_record.plot(ax=ax, with_ruler=False, strand_in_label_threshold=2)

        self.fig = fig



from datetime import datetime
import os


try:
    import pdf_reports.tools as pdf_tools
    from pdf_reports import (
        add_css_class,
        dataframe_to_html,
        pug_to_html,
        style_table_rows,
        write_report,
    )
    import pandas
    import matplotlib

    REPORT_PKGS_AVAILABLE = True
except ImportError:
    REPORT_PKGS_AVAILABLE = False

from importlib.metadata import version

version_number = version("plasmid_assessor")

THIS_PATH = os.path.dirname(os.path.realpath(__file__))
ASSETS_PATH = os.path.join(THIS_PATH, "report_assets")
REPORT_TEMPLATE = os.path.join(ASSETS_PATH, "assessment_report.pug")
STYLESHEET = os.path.join(ASSETS_PATH, "report_style.css")


def end_pug_to_html(template, **context):
    now = datetime.now().strftime("%Y-%m-%d")
    defaults = {
        "sidebar_text": "Generated on %s by EGF's Plasmid assessor (version %s)"
        % (now, version_number),
        "plasma_logo_url": os.path.join(ASSETS_PATH, "imgs", "plasma_logo.png"),
    }
    for k in defaults:
        if k not in context:
            context[k] = defaults[k]
    return pug_to_html(template, **context)


def write_pdf_report(target, assessment):
    """Write an assessment report with a PDF summary.

    Parameters
    ----------
    target : str
        Path for PDF file.
    assessment : Assessment
        Assessment instance.
    """
    if not REPORT_PKGS_AVAILABLE:
        raise ImportError(
            "Install extra packages with `pip install plasmid_assessor[report]`"
        )

    assessment.figure_data = pdf_tools.figure_data(assessment.fig, fmt="svg")

    html = end_pug_to_html(
        REPORT_TEMPLATE,
        assessment=assessment,
    )
    write_report(html, target, extra_stylesheets=(STYLESHEET,))



import Bio.Restriction
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import plasmid_assessor as plasma


def test_assess_seq():
    # Also tests digest_seq()
    record = SeqRecord(
        Seq("AAAAACGTCTCAACTG" + "AAAAA" + "TATCAGAGACGAAAAA"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(record, "BsmBI")
    design.assess_seq()
    assert design.results["digest"]["left_overhang"] == "ACTG"
    assert design.results["digest"]["right_overhang"] == "TATC"


def test_get_number_of_sites():
    sequence = SeqRecord(
        Seq("ATCGATCG"), annotations={"topology": "circular"}
    )  # 0 site
    design = plasma.Assessment(sequence, "BsmBI")
    design.get_number_of_sites()
    assert design.results["number_of_sites"] == 0

    sequence = SeqRecord(
        Seq("AAAAACGTCTCAACTGAAAAAATATCAGAGACGAAAAA"),
        annotations={"topology": "circular"},
    )  # 2 sites
    design = plasma.Assessment(sequence, "BsmBI")
    design.get_number_of_sites()
    assert design.results["number_of_sites"] == 2


def test_evaluate_orientation():
    sequence = SeqRecord(
        Seq("AAAA" + "CGTCTCAACTG" + "AAAAA" + "TATCAGAGACG" + "AAAA"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(sequence, "BsmBI")
    design.evaluate_orientation()
    assert design.results["is_site_orientation_correct"]


def test_count_other_sites():
    sequence = SeqRecord(
        Seq("CGTCTCAACTG" + "AAA" + "TATCAGAGACG" + "AGGTCTC"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(sequence, "BsmBI")
    design.count_other_sites(other_enzymes=["BsaI"])
    assert design.results["other_sites"]["has_any_other_sites"]
    assert len(design.results["other_sites"]["enzyme"][Bio.Restriction.BsaI]) == 1


def test_linear_seq():
    record = SeqRecord(
        Seq("AAAAA" + "CGTCTCAACTG" + "AAAAA" + "TATCAGAGACG" + "AAAAA"),
        annotations={"topology": "linear"},
    )
    design = plasma.Assessment(record, "BsmBI")
    design.assess_seq()


def test_circular_seq():
    # The retained insert (part) is one continuous sequence
    record = SeqRecord(
        Seq("AAAAA" + "CGTCTCAACTG" + "AAAAA" + "TATCAGAGACG" + "AAAAA"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(record, "BsmBI")
    design.assess_seq()


def test_circular_seq_origin_in_insert():
    # In this version, the retained insert is split to the start and end of the seq
    record = SeqRecord(
        Seq("AAAAA" + "TATCAGAGACG" + "AAAAA" + "CGTCTCAACTG" + "AAAAA"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(record, "BsmBI")
    design.assess_seq()


def test_plot_sequence():
    sequence = SeqRecord(
        Seq("CGTCTCAACTG" + "AAA" + "TATCAGAGACG" + "AGGTCTC"),
        annotations={"topology": "circular"},
    )
    design = plasma.Assessment(sequence, "BsmBI")
    design.plot_sequence()
    assert design.fig
